!+
! Module: cal_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Returns semi-axes of an ellipse round each of a list of sources.
! 
! Comments: TODO:
! 
! $Id: cal_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module cal_aux

  use dal
  use cal
  use oal
  use errorHandling

  use constants

  implicit none 

  private 

  public getCalInfo
  public getInstrumentAttitude
  public getScAttitude

  interface getCalInfo
    module procedure getCalInfo_name
    module procedure getCalInfo_set
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine getCalInfo_name(setName, instrumentId, expStartTimeStamp&
  , scAttitude, instrumAttitude)
  !---------------------------------------------------------------------
  ! A routine to extract (directly or indirectly) the instrumentId,
  ! expStartTimeStamp, scAttitude and instrumAttitude from the header
  ! of the specified data set.
  !
  ! NOTE! This routine also sets the instrument state of the cal.

    implicit none

    character(*), intent(in) :: setName
    integer(int32), intent(out) ::&
      instrumentId
    real(double), intent(out) ::&
      expStartTimeStamp
    type(SpacecraftAttitudeType), intent(out) ::&
      scAttitude, instrumAttitude

    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    call getCalInfo_set(set, instrumentId, expStartTimeStamp, scAttitude&
    , instrumAttitude)
    call release(set)

  end subroutine getCalInfo_name

  !---------------------------------------------------------------------
  subroutine getCalInfo_set(set, instrumentId, expStartTimeStamp, scAttitude&
  , instrumAttitude)
  !---------------------------------------------------------------------
  ! A routine to extract (directly or indirectly) the instrumentId,
  ! expStartTimeStamp, scAttitude and instrumAttitude from the header
  ! of the specified data set.
  !
  ! NOTE! This routine also sets the instrument state of the cal.

    implicit none

    type(DataSetT), intent(in) :: set
    integer(int32), intent(out) ::&
      instrumentId
    real(double), intent(out) ::&
      expStartTimeStamp
    type(SpacecraftAttitudeType), intent(out) ::&
      scAttitude, instrumAttitude

    character(20) :: instrumentName, dateTimeStr
    real(double) ::&
      pntRaDeg,&
      pntDecDeg,&
      pntAposDeg

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Noisy, 'ssclib: retrieve instrum, dateobs, attitude.')

    instrumentName = stringAttribute(set, "INSTRUME")
    instrumentId = instId(instrumentName)

    if (instrumentId .ne. EMOS1&
    .and. instrumentId .ne. EMOS2&
    .and. instrumentId .ne. EPN) then
call error('badInstrum', 'Instrument '//trim(instrumentName)//' not supported.')
    end if

    pntRaDeg   = real64Attribute(set, "RA_PNT")
    pntDecDeg  = real64Attribute(set, "DEC_PNT")
    pntAposDeg = real64Attribute(set, "PA_PNT")

    scAttitude%ra   = pntRaDeg   * piOn180
    scAttitude%dec  = pntDecDeg  * piOn180
    scAttitude%apos = pntAposDeg * piOn180

!******* It actually SHOULD be the INSTRUMENT attitude, but was still s/c, as of 1 Feb 2002, in both MOS and PN event lists. When the change comes along we will need to read RA_SCX, _SCY etc kwds to get the s/c attitude. Handy to have a CAL call to convert SCX/Y/Z to scAttitude... Uwe??

    dateTimeStr = stringAttribute(set, "DATE-OBS")
    expStartTimeStamp = toTimeTag(dateTimeStr) ! in caloalutils/caloaldefs.f90

    call CAL_setState(instrument=instrumentId, randomize=.false.)
    instrumAttitude = getInstrumentAttitude(scAttitude, expStartTimeStamp)

  end subroutine getCalInfo_set

  !---------------------------------------------------------------------
  function getInstrumentAttitude(scAttitude, timeStamp)&
  result(instrumAttitude)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the
  ! calling routine.

    implicit none

    real(double), intent(in) ::&
      timeStamp
    type(SpacecraftAttitudeType), intent(in) ::&
      scAttitude ! typedef from caloalutils/src/caloaldefs.f90
    type(SpacecraftAttitudeType) ::&
      instrumAttitude ! typedef from caloalutils/src/caloaldefs.f90

    real(double), dimension(:,:), pointer ::&
      boresightMatrix
    real(double), dimension(3,3) ::&
      scAttitudeMatrix, instrumAttitudeMatrix
    type(EulerAnglesType) ::&
      instrumentEulerAngles ! typedef from oal/f90/oaltypes.f90
!real(double) :: x, y, z

    !-------------------------------------------------------------------
    ! Start

    call OAL_toAttitudeMatrix(scAttitude, scAttitudeMatrix)
    !
    ! The array returned by this call represents the spacecraft attitude
    ! matrix ASC_sky which has, in its ith row, a vector of direction
    ! cosines in the sky basis of the ith axis of the spacecraft
    ! reference frame (the same frame as defines the SACCOORD system).
    ! For example, the x or 1st axis of this frame (which points
    ! approximately in the opposite direction to the star tracker and
    ! the optical axes of the instruments) appears in the first row of
    ! ASC_sky. The sky basis is that right-handed frame which has its
    ! x and y axes at declination=0, z axis pointing in the declination
    ! = pi/2 direction, x in the ra=0 and y in the ra=pi/2 (ie, 6 hrs)
    ! directions.
    !
    ! This matrix can be used to convert a vector from the sky basis to
    ! the spacecraft or SACCOORD basis, as follows:
    !
    !		v_sc = ASC_sky * v_sky.
    !
    ! NOTE! The second index of the array scAttitudeMatrix corresponds
    ! to the row number of ASC_sky. scAttitudeMatrix should therefore be
    ! transposed in order to conform to the normal convention in which
    ! the first index corresponds to row number.
    !
    ! So:

    scAttitudeMatrix = transpose(scAttitudeMatrix)

    call CAL_getBoresightMatrix(timeStamp, boresightMatrix)
    !
    ! The boresight matrix AIN_sc is defined as follows:
    !
    !	AIN_sky = AIN_sc * ASC_sky
    !
    ! where the ith row of the generic matrix AIN_basis is the vector of
    ! direction cosines (in the given basis) of the ith axis of the
    ! instrument frame. The instrument frame defines the TELCOORD system:
    ! the essential point is that its x axis is aligned with the optical
    ! axis of the instrument, roughly in the opposite direction to the
    ! x axis of the spacecraft or SACCOORD frame.
    !
    ! The boresight matrix AIN_sc can therefore be seen to be just the
    ! instrument attitude matrix, but expressed in a basis defined by
    ! the spacecraft frame.
    !
    ! NOTE!! The call CAL_getBoresightMatrix, in contrast to most of
    ! the other cal and oal calls that return an attitude matrix,
    ! returns an array in the CORRECT convention: that is, the first
    ! index of the array boresightMatrix as returned above represents
    ! the row index of the associated matrix AIN_sc. Thus it should NOT be
    ! transposed before use.

    instrumAttitudeMatrix = matmul(boresightMatrix, scAttitudeMatrix)
    !
    ! This array is still the correct convention, ie the first index of
    ! the array represents the row index of the associated matrix AIN_sky.

!x = instrumAttitudeMatrix(1,1)
!y = instrumAttitudeMatrix(1,2)
!z = instrumAttitudeMatrix(1,3)
!if (x .ne. 0.0 .and. y .ne. 0.0) then
!  instrumAttitude%ra   =  atan2(y, x) 
!else
!  instrumAttitude%ra   =  0.0 
!end if
!instrumAttitude%dec  = asin(z)

    call OAL_toEulerAngles(transpose(instrumAttitudeMatrix)&
    , instrumentEulerAngles)
    instrumAttitude%ra   =  instrumentEulerAngles%phi 
    instrumAttitude%dec  = -instrumentEulerAngles%theta
    instrumAttitude%apos = -instrumentEulerAngles%psi
    !
    ! Don't ask me why this works! But it does. If you're worried about
    ! it then use the commented-out xyz stuff immediately above! That
    ! works too and it is easier to follow why.

    do while(instrumAttitude%ra .lt. 0.0)
      instrumAttitude%ra = instrumAttitude%ra + TwoPi
    end do
    do while(instrumAttitude%apos .lt. 0.0)
      instrumAttitude%apos = instrumAttitude%apos + TwoPi
    end do

    call CAL_releaseMemory(boresightMatrix)

  end function getInstrumentAttitude

  !---------------------------------------------------------------------
  function getScAttitude(instrumAttitude, timeStamp)&
  result(scAttitude)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the
  ! calling routine.

    implicit none

    real(double), intent(in) ::&
      timeStamp
    type(SpacecraftAttitudeType), intent(in) ::&
      instrumAttitude ! typedef from caloalutils/src/caloaldefs.f90
    type(SpacecraftAttitudeType) ::&
      scAttitude ! typedef from caloalutils/src/caloaldefs.f90

    real(double), dimension(:,:), pointer ::&
      boresightMatrix
    real(double), dimension(3,3) ::&
      scAttitudeMatrix, instrumAttitudeMatrix
    type(EulerAnglesType) ::&
      scEulerAngles ! typedef from oal/f90/oaltypes.f90
!real(double) :: x, y, z

    !-------------------------------------------------------------------
    ! Start

    call OAL_toAttitudeMatrix(instrumAttitude, instrumAttitudeMatrix)
    !
    ! The array returned by this call represents the instrument attitude
    ! matrix AIN_sky which has, in its ith row, a vector of direction
    ! cosines in the sky basis of the ith axis of the instrument
    ! reference frame (the same frame as defines the TELCOORD system).
    ! For example, the x or 1st axis of this frame (which points
    ! approximately in the same direction to the star tracker and
    ! the optical axes of the instruments) appears in the first row of
    ! AIN_sky. The sky basis is that right-handed frame which has its
    ! x and y axes at declination=0, z axis pointing in the declination
    ! = pi/2 direction, x in the ra=0 and y in the ra=pi/2 (ie, 6 hrs)
    ! directions.
    !
    ! This matrix can be used to convert a vector from the sky basis to
    ! the instrument or TELCOORD basis, as follows:
    !
    !		v_ins = Ain_sky * v_sky.		(1)
    !
    ! NOTE! The second index of the array instrumAttitudeMatrix corresponds
    ! to the row number of AIN_sky. instrumAttitudeMatrix should therefore
    ! be transposed in order to conform to the normal convention in which
    ! the first index corresponds to row number.
    !
    ! So:

    instrumAttitudeMatrix = transpose(instrumAttitudeMatrix)

    call CAL_getBoresightMatrix(timeStamp, boresightMatrix)
    !
    ! The boresight matrix AIN_sc is defined as follows:
    !
    !		AIN_sky = AIN_sc * ASC_sky		(2)
    !
    ! where the ith row of the generic matrix AIN_basis is the vector of
    ! direction cosines (in the given basis) of the ith axis of the
    ! instrument frame. The boresight matrix AIN_sc can therefore be
    ! seen to be just the instrument attitude matrix, but expressed in
    ! a basis defined by the spacecraft frame.
    !
    ! The spacecraft frame is contained in the matrix ASC_sky which has,
    ! in its ith row, a vector of
    ! direction cosines in the sky basis of the ith axis of the spacecraft
    ! reference frame (the same frame as defines the SACCOORD system).
    ! For example, the x or 1st axis of this frame (which points
    ! approximately in the opposite direction to the star tracker and
    ! the optical axes of the instruments) appears in the first row of
    ! ASC_sky.
    !
    ! NOTE!! The call CAL_getBoresightMatrix, in contrast to most of
    ! the other cal and oal calls that return an attitude matrix,
    ! returns an array in the CORRECT convention: that is, the first
    ! index of the array boresightMatrix as returned above represents
    ! the row index of the associated matrix AIN_sc. Thus it should NOT be
    ! transposed before use in equation (2) above. However, don't be
    ! confused! In the present subroutine I'm about to transpose it
    ! anyway. What gives? Well, to extract ASC_sky we need to
    ! premultiply equation (2) by the inverse of AIN_sc. But for
    ! attitude (indeed, any orthogonal) matrices, the inverse is just
    ! equal to the transpose. Hence
    !
    !		ASC_sky = AIN_sc^T * AIN_sky		(3)

    boresightMatrix = transpose(boresightMatrix)
    scAttitudeMatrix = matmul(boresightMatrix, instrumAttitudeMatrix)
    !
    ! The resulting array scAttitudeMatrix is still the correct
    ! convention, ie the first index of the array represents the row
    ! index of the associated matrix AIN_sky.

!x = scAttitudeMatrix(1,1)
!y = scAttitudeMatrix(1,2)
!z = scAttitudeMatrix(1,3)
!if (x .ne. 0.0 .and. y .ne. 0.0) then
!  scAttitude%ra   =  atan2(y, x) 
!else
!  scAttitude%ra   =  0.0 
!end if
!scAttitude%dec  = asin(z)

    call OAL_toEulerAngles(transpose(scAttitudeMatrix)&
    , scEulerAngles)
    scAttitude%ra   =  scEulerAngles%phi 
    scAttitude%dec  = -scEulerAngles%theta
    scAttitude%apos = -scEulerAngles%psi
    !
    ! Don't ask me why this works! But it does. If you're worried about
    ! it then use the commented-out xyz stuff immediately above! That
    ! works too and it is easier to follow why.

    do while(scAttitude%ra .lt. 0.0)
      scAttitude%ra = scAttitude%ra + TwoPi
    end do
    do while(scAttitude%apos .lt. 0.0)
      scAttitude%apos = scAttitude%apos + TwoPi
    end do

    call CAL_releaseMemory(boresightMatrix)

  end function getScAttitude

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module cal_aux
