!+
! Module: source_cutouts
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Returns semi-axes of an ellipse round each of a list of sources.
! 
! Comments: TODO:
! 
! $Id: source_cutouts.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module source_cutouts

  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use dal
  use errorHandling

  use coordinate
  use my_psf
  use mathutils
  use wcs_aux
  use constants
  use fits_utils

  implicit none 

  private 

  public getSourceCutouts

contains 

  !---------------------------------------------------------------------
  subroutine getSourceCutouts(sourceTab, imgWcs, srcWcs, scAttitude&
  , expStartTimeStamp, fracBackground, raColName, decColName, rateColName&
  , bkgFluxColName, xPixels, yPixels, ellipseXampPixels, ellipseYampPixels&
  , ellipsePhase, srcIsValid)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up with the correct instrument beforehand.

    implicit none

  type :: localSrcDataType
    real(single) ::&
      bkgFlux,&		! in counts per square arc second
      counts,&
      axisLongWorld,&
      axisShrtWorld,&
      rotatedAngleWorld!!,&
!      rate
    real(double) ::&
      raDeg,&
      decDeg,&
      EeV,&
      thetaArcsec,&
      phi,&
      xWorld,&
      yWorld,&
      fovDistanceWorld
  end type localSrcDataType

    type(TableT), intent(in) :: sourceTab
    type(wcsDataType), intent(in) :: imgWcs, srcWcs
    real(double), intent(in) ::&
      expStartTimeStamp
    type(SpacecraftAttitudeType), intent(in) ::&
      scAttitude
    real(single), intent(in) ::&
      fracBackground
    character(*), intent(in) :: raColName, decColName, rateColName,&
      bkgFluxColName
    real(single), pointer ::&
      xPixels(:), yPixels(:), ellipseXampPixels(:), ellipseYampPixels(:),&
      ellipsePhase(:)
    logical, pointer ::&
      srcIsValid(:)

    integer(int32) :: numSources, src
    real(single), dimension(:), allocatable ::&
      bkgFluxPixels,&	! in counts (per sec) per image pixel
      bkgFluxArcsec,&	! in counts (per sec) per square arc second
      ellipseLongAxisArcsec,&
      ellipseShrtAxisArcsec,&
      ellipseLongAxisTan,&
      ellipseShrtAxisTan,&
      ellipseRotAngleDeg,&
      ellipseXampTan, ellipseYampTan,&
      rate,&
      detX, detY,&
      peakFluxArcsec	! in counts (per sec) per square arc second
    real(double), dimension(:), allocatable ::&
      raDeg,&
      decDeg,&
      EeV,&
      thetaArcsec,&
      phi,&
      xTan,&
      yTan
    real(single) ::&
      averageBkgRatePixels, srcPixelSizeArcsec
    real(double) ::&
      refRa, refDec, ctrFovXtan, ctrFovYtan
    integer :: foreignStatus
logical :: testFlag = .false.
real :: ctrFovXpixels, ctrFovYpixels
!real(double) :: ctrFovRaDeg, ctrFovDecDeg

    !-------------------------------------------------------------------
    ! Start

!.......................................................................
! Load source data from the table:

    numSources = numberOfRows(sourceTab)

    ! Pointers returned by the subroutine, which should be deallocated by
    ! the calling routine after use:
    allocate(xPixels(numSources), yPixels(numSources))
    allocate(ellipseXampPixels(numSources))
    allocate(ellipseYampPixels(numSources))
    allocate(ellipsePhase(numSources), srcIsValid(numSources))

    ! Other source data vectors:
    allocate(bkgFluxPixels(numSources), bkgFluxArcsec(numSources))
    allocate(ellipseLongAxisArcsec(numSources))
    allocate(ellipseShrtAxisArcsec(numSources))
    allocate(ellipseLongAxisTan(numSources), ellipseShrtAxisTan(numSources))
    allocate(ellipseRotAngleDeg(numSources))
    allocate(ellipseXampTan(numSources), ellipseYampTan(numSources))
    allocate(rate(numSources), peakFluxArcsec(numSources))
    allocate(detX(numSources), detY(numSources))
    allocate(raDeg(numSources), decDeg(numSources))
    allocate(EeV(numSources))
    allocate(thetaArcsec(numSources), phi(numSources))
    allocate(xTan(numSources), yTan(numSources))

    call readColData(sourceTab, raColName,      raDeg)
    call readColData(sourceTab, decColName,     decDeg)
    call readColData(sourceTab, rateColName,    rate)
    call readColData(sourceTab, bkgFluxColName, bkgFluxPixels)

if (testFlag) then
do src = 1, size(raDeg)
  print *, 'src, rate(src), bkgFluxPixels(src): '&
  , src, rate(src), bkgFluxPixels(src)
end do
print *, ''
end if

    ! Note: rate and bkgFlux ought to be in compatible units, eg:
    !	rate:		total source count rate (ie counts / sec)
    !	bkgFluxPixels:	bkg counts / sec / pixel
    ! or:
    !	rate:		total source counts
    !	bkgFluxPixels:	bkg counts / pixel
    ! Either way, bkgFluxPixels must be counts per pixels of the image used
    ! originally for source detection. The size of these pixels must be
    ! obtained in order to convert the bkgFluxPixels to counts (per sec) per
    ! square arcsec.

!.......................................................................
! Do some checks:

    if (all(rate .le. 0.0)) then
      call warning('allSrcNonPositive', 'None of the sources have values &
      &of '//trim(rateColName)//' greater than zero.')
  return
    end if

    if (all(rate .le. 0.0 .or. bkgFluxPixels .le. 0.0)) then
      call warning('allSrcBkgNonPositive', 'None of the positive-flux &
      &sources have a positive-valued local background '&
      //trim(bkgFluxColName)//'.')
  return
    end if

    if (any(rate .gt. 0.0 .and. bkgFluxPixels .le. 0.0)) then
      call warning('nonPositiveSrcBkg', 'Some of the '//trim(bkgFluxColName)&
      //' values of the source list are less than or equal to zero. These &
      &have been set to the average of the positive values.')

      averageBkgRatePixels = sum(bkgFluxPixels, (rate .gt. 0.0&
      .and. bkgFluxPixels .gt. 0.0)) / count(rate .gt. 0.0&
      .and. bkgFluxPixels .gt. 0.0)

      where(rate .gt. 0.0 .and. bkgFluxPixels .le. 0.0)&
        bkgFluxPixels = averageBkgRatePixels
   end if

!.......................................................................
! Correct bkg flux for pixel size:

   srcPixelSizeArcsec = abs(srcWcs%X%pixelSizeWorld * srcWcs%Y%pixelSizeWorld)&
     * 3600.0 * 3600.0

   bkgFluxArcsec = bkgFluxPixels / srcPixelSizeArcsec

!.......................................................................
! Get src fov coords:

    call raDecToDetXY(raDeg, decDeg, scAttitude, expStartTimeStamp&
    , detX, detY, thetaArcsec, phi, foreignStatus)
    ! in new_coordinate.

    if (foreignStatus == -1) then
call error('bug1', 'Contact the code developer.')
    else if (foreignStatus < 0) then
call error('bug2', 'Contact the code developer.')
    end if

!.......................................................................
! Get peak fluxes:

EeV = 500.0d0
!***** temporary bodge.

    do src = 1, numSources
      peakFluxArcsec(src) = rate(src)&
      / getPsfIntegral(thetaArcsec(src), EeV(src))
      ! in counts per square arc second
    end do

if (testFlag) then
do src = 1, size(raDeg)
  print *, 'src, bkgFluxArcsec(src), peakFluxArcsec(src): '&
  , src, bkgFluxArcsec(src), peakFluxArcsec(src)
end do
print *, ''
end if

!.......................................................................
! Get ellipse axes:

    call getPsfEllipseAxes(thetaArcsec, EeV, peakFluxArcsec&
      , bkgFluxArcsec, fracBackground, ellipseLongAxisArcsec&
      , ellipseShrtAxisArcsec, srcIsValid)

!**** check and report about invalid sources?

if (testFlag) then
do src = 1, size(raDeg)
  print *, 'src, ellipseShrtAxisArcsec(src), ellipseLongAxisArcsec(src), srcIsValid(src): '&
  , src, ellipseShrtAxisArcsec(src), ellipseLongAxisArcsec(src), srcIsValid(src)
end do
print *, ''
end if

!.......................................................................
! Convert ra and dec to image tangent coords:

!    if (imgWcsType .eq. 'DETXY') then
!      srcData%xWorld = srcXdet
!      srcData%yWorld = srcYdet
!
!      srcData%axisLongWorld = axisLongArcsec / detUnitArcsec
!      srcData%axisShrtWorld = axisShrtArcsec / detUnitArcsec
!
!      fovCtrXworld = 0.0
!      fovCtrYworld = 0.0
!
!    else if (imgWcsType .eq. 'XY') then

      refRa  = imgWcs%X%refWorld * PiOn180
      refDec = imgWcs%Y%refWorld * PiOn180

      call raDecToTan(refRa, refDec, raDeg * PiOn180, decDeg * PiOn180&
        , xTan, yTan, foreignStatus)
      ! in new_coordinate.

!*** check foreignStatus

if (testFlag) then
do src = 1, size(raDeg)
  print *, 'src, raDeg(src), decDeg(src), xTan(src), yTan(src): '&
  , src, raDeg(src), decDeg(src), xTan(src), yTan(src)
end do
print *, ''
end if

      call detXYtoTan(ctrFovXtan, ctrFovYtan, refRa, refDec, scAttitude&
      , expStartTimeStamp, thetaArcsec=0.0d0, phi=0.0d0, status=foreignStatus)

!call detXYtoRaDec(ctrFovRaDeg, ctrFovDecDeg, scAttitude&
!, expStartTimeStamp, thetaArcsec=0.0d0, phi=0.0d0, status=foreignStatus)
!
!if (testFlag) then
!  print *, 'ctrFovRaDeg, ctrFovDecDeg: ', ctrFovRaDeg, ctrFovDecDeg
!end if
!
!call raDecToTan(refRa, refDec, ctrFovRaDeg*PiOn180, ctrFovDecDeg*PiOn180, ctrFovXtan, ctrFovYtan, status=foreignStatus)

!*** check foreignStatus

      ellipseShrtAxisTan = ellipseShrtAxisArcsec / 3600.0 * PiOn180
      ellipseLongAxisTan = ellipseLongAxisArcsec / 3600.0 * PiOn180
      ellipseRotAngleDeg = atan2(yTan - ctrFovYtan, xTan - ctrFovXtan) / PiOn180

if (testFlag) then
do src = 1, size(raDeg)
  print *, 'src, ellipseShrtAxisTan(src), ellipseLongAxisTan(src), ellipseRotAngleDeg(src): '&
  , src, ellipseShrtAxisTan(src), ellipseLongAxisTan(src), ellipseRotAngleDeg(src)
end do
print *, ''
end if

!.......................................................................
! Convert ellipses from axes and rotAngleDeg to amps and phase:

      call ellipseAngleToPhase(ellipseShrtAxisTan, ellipseLongAxisTan&
        , ellipseRotAngleDeg, ellipseXampTan, ellipseYampTan, ellipsePhase)

if (testFlag) then
do src = 1, size(raDeg)
  print *, 'src, ellipseXampTan(src), ellipseYampTan(src), ellipsePhase(src): '&
  , src, ellipseXampTan(src), ellipseYampTan(src), ellipsePhase(src)
end do
print *, ''
end if

!.......................................................................
! Change image tangents to pixel coords, also scale ellipse amps to
! pixel coords:

      xPixels = xTan / PiOn180 / abs(imgWcs%X%pixelSizeWorld) + imgWcs%X%refPixels
      yPixels = yTan / PiOn180 / abs(imgWcs%Y%pixelSizeWorld) + imgWcs%Y%refPixels
      ellipseXampPixels = ellipseXampTan / PiOn180 / abs(imgWcs%X%pixelSizeWorld)
      ellipseYampPixels = ellipseYampTan / PiOn180 / abs(imgWcs%Y%pixelSizeWorld)

ctrFovXpixels = ctrFovXtan / PiOn180 / abs(imgWcs%X%pixelSizeWorld) + imgWcs%X%refPixels
ctrFovYpixels = ctrFovYtan / PiOn180 / abs(imgWcs%Y%pixelSizeWorld) + imgWcs%Y%refPixels

if (testFlag) then
print *, 'ctrFovXpixels, ctrFovYpixels: '&
, ctrFovXpixels, ctrFovYpixels
print *, ''
call pgpt1(ctrFovXpixels, ctrFovYpixels, 2)
do src = 1, size(raDeg)
  print *, 'src, ellipseXampPixels(src), ellipseYampPixels(src): '&
  , src, ellipseXampPixels(src), ellipseYampPixels(src)
end do
print *, ''
end if

!      do src = 1, numSources
!        call projectToXY(refRa, refDec, srcData(src)%raDeg * PiOn180&
!        , srcData(src)%decDeg * PiOn180, srcX, srcY)
!
!        srcData(src)%xWorld = imgWcs%X%refWorld - srcX / PiOn180
!        srcData(src)%yWorld = imgWcs%Y%refWorld + srcY / PiOn180
!      end do ! src
!
!      srcData%axisLongWorld = axisLongArcsec / 3600.0
!      srcData%axisShrtWorld = axisShrtArcsec / 3600.0
!
!      call projectToXY(refRa, refDec, instrumAttitude%ra&
!      , instrumAttitude%dec, srcX, srcY)
!
!      fovCtrXworld = imgWcs%X%refWorld - srcX / PiOn180
!      fovCtrYworld = imgWcs%Y%refWorld + srcY / PiOn180

!    else
!call error('badWcsType', "Don't recognise WCS style.")
!    end if

!!.......................................................................
!! Calculate the rotated angle of the ellipse:
!
!    call message(AppMsg, Verbose, "Last bits and pieces.")
!
!    dx = srcData%xWorld - fovCtrXworld
!    dy = srcData%yWorld - fovCtrYworld
!
!    do src = 1, numSources
!      srcData(src)%rotatedAngleWorld = atan2(dy(src), dx(src))
!    end do
!
!    srcData%fovDistanceWorld = sqrt(dx * dx + dy * dy)

!    if (present(status)) status = localStatus

  end subroutine getSourceCutouts

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module source_cutouts
